@@ -1,59 +1,144 @@
 """ The access backend object base class """
+import hashlib
+import hmac
+import time
 from collections import defaultdict
-from passlib.apps import custom_app_context as pwd_context
-from pyramid.security import (Authenticated, Everyone,
-                              effective_principals, Allow, Deny,
-                              ALL_PERMISSIONS)
+from typing import Any, Dict, List, Optional, Set, Tuple
+
+from passlib.apps import LazyCryptContext
+from passlib.utils import sys_bits
+from pyramid.security import ALL_PERMISSIONS, Allow, Authenticated, Deny, Everyone
 from pyramid.settings import aslist
 
+from pypicloud.util import EnvironSettings
 
-def group_to_principal(group):
-    """ Convert a group to its corresponding principal """
-    if group in (Everyone, Authenticated) or group.startswith('group:'):
+Admin = "admin"
+
+# Roughly tuned using https://bitbucket.org/ecollins/passlib/raw/default/choose_rounds.py
+# For 10ms. This differs from the passlib recommendation of 350ms due to the difference in use case
+DEFAULT_ROUNDS = {
+    "sha512_crypt__default_rounds": 20500,
+    "sha256_crypt__default_rounds": 16000,
+    "pbkdf2_sha512__default_rounds": 10500,
+    "pbkdf2_sha256__default_rounds": 13000,
+    "bcrypt__default_rounds": 7,
+    "argon2__default_rounds": 1,
+}
+SCHEMES = [
+    "sha512_crypt",
+    "sha256_crypt",
+    "pbkdf2_sha512",
+    "pbkdf2_sha256",
+    "bcrypt",
+    "argon2",
+]
+if sys_bits < 64:
+    SCHEMES.remove("sha512_crypt")
+    SCHEMES.remove("pbkdf2_sha512")
+
+
+def get_pwd_context(
+    preferred_hash: Optional[str] = None, rounds: Optional[int] = None
+) -> LazyCryptContext:
+    """Create a passlib context for hashing passwords"""
+    if preferred_hash is None or preferred_hash == "sha":
+        preferred_hash = "sha256_crypt" if sys_bits < 64 else "sha512_crypt"
+    if preferred_hash == "pbkdf2":
+        preferred_hash = "pbkdf2_sha256" if sys_bits < 64 else "pbkdf2_sha512"
+
+    if preferred_hash not in SCHEMES:
+        raise Exception(
+            "Password hash %r is not in the list of supported schemes" % preferred_hash
+        )
+
+    # Put the preferred hash at the beginning of the schemes list
+    schemes = list(SCHEMES)
+    schemes.remove(preferred_hash)
+    schemes.insert(0, preferred_hash)
+
+    # Override the default rounds of the preferred hash, if provided
+    default_rounds = dict(DEFAULT_ROUNDS)
+    if rounds is not None:
+        default_rounds[preferred_hash + "__default_rounds"] = rounds
+
+    return LazyCryptContext(schemes=schemes, default=schemes[0], **default_rounds)
+
+
+def group_to_principal(group: str) -> str:
+    """Convert a group to its corresponding principal"""
+    if group in (Everyone, Authenticated) or group.startswith("group:"):
         return group
-    elif group == 'everyone':
+    elif group == "everyone":
         return Everyone
-    elif group == 'authenticated':
+    elif group == "authenticated":
         return Authenticated
     else:
-        return 'group:' + group
+        return "group:" + group
 
 
-def groups_to_principals(groups):
-    """ Convert a list of groups to a list of principals """
+def groups_to_principals(groups: List[str]) -> List[str]:
+    """Convert a list of groups to a list of principals"""
     return [group_to_principal(g) for g in groups]
 
 
+ONE_WEEK = 60 * 60 * 24 * 7
+
+
 class IAccessBackend(object):
 
-    """ Base class for retrieving user and package permission data """
+    """Base class for retrieving user and package permission data"""
 
     mutable = False
     ROOT_ACL = [
-        (Allow, Authenticated, 'login'),
-        (Allow, 'admin', ALL_PERMISSIONS),
+        (Allow, Authenticated, "login"),
+        (Allow, Admin, ALL_PERMISSIONS),
         (Deny, Everyone, ALL_PERMISSIONS),
     ]
 
-    def __init__(self, request=None, default_read=None, default_write=None,
-                 cache_update=None):
+    def __init__(
+        self,
+        request=None,
+        default_read=None,
+        default_write=None,
+        disallow_fallback=(),
+        cache_update=None,
+        pwd_context=None,
+        token_expiration=ONE_WEEK,
+        signing_key=None,
+    ):
         self.request = request
         self.default_read = default_read
         self.default_write = default_write
+        self.disallow_fallback = disallow_fallback
         self.cache_update = cache_update
+        self.pwd_context = pwd_context
+        self.token_expiration = token_expiration
+        self.signing_key = signing_key
 
     @classmethod
-    def configure(cls, settings):
-        """ Configure the access backend with app settings """
+    def configure(cls, settings: EnvironSettings) -> Dict[str, Any]:
+        """Configure the access backend with app settings"""
+        rounds = settings.get("auth.rounds")
+        scheme = settings.get("auth.scheme")
         return {
-            'default_read': aslist(settings.get('pypi.default_read',
-                                                ['authenticated'])),
-            'default_write': aslist(settings.get('pypi.default_write', [])),
-            'cache_update': aslist(settings.get('pypi.cache_update',
-                                                ['authenticated'])),
+            "default_read": aslist(
+                settings.get("pypi.default_read", ["authenticated"])
+            ),
+            "default_write": aslist(settings.get("pypi.default_write", [])),
+            "disallow_fallback": aslist(settings.get("pypi.disallow_fallback", [])),
+            "cache_update": aslist(
+                settings.get("pypi.cache_update", ["authenticated"])
+            ),
+            "pwd_context": get_pwd_context(scheme, rounds),
+            "token_expiration": int(settings.get("auth.token_expire", ONE_WEEK)),
+            "signing_key": settings.get("auth.signing_key"),
         }
 
-    def allowed_permissions(self, package):
+    @classmethod
+    def postfork(cls, **kwargs):
+        """This method will be called after uWSGI forks"""
+
+    def allowed_permissions(self, package: str) -> Dict[str, Tuple[str, ...]]:
         """
         Get all allowed permissions for all principals on a package
 
@@ -64,46 +149,51 @@ class IAccessBackend(object):
 
         """
         all_perms = {}
-        for user, perms in self.user_permissions(package).iteritems():
-            all_perms['user:' + user] = tuple(perms)
+        for user, perms in self.user_permissions(package).items():
+            all_perms["user:" + user] = tuple(perms)
 
-        for group, perms in self.group_permissions(package).iteritems():
+        for group, perms in self.group_permissions(package).items():
             all_perms[group_to_principal(group)] = tuple(perms)
 
         # If there are no group or user specifications for the package, use the
         # default
-        if len(all_perms) == 0:
+        if not all_perms:
             for principal in groups_to_principals(self.default_read):
-                all_perms[principal] = ('read',)
+                all_perms[principal] = ("read",)
             for principal in groups_to_principals(self.default_write):
                 if principal in all_perms:
-                    all_perms[principal] += ('write',)
+                    all_perms[principal] += ("write",)
                 else:
-                    all_perms[principal] = ('write',)
+                    all_perms[principal] = ("write",)
+
+        # add fallback permissions
+        if package not in self.disallow_fallback:
+            for principal in all_perms:
+                all_perms[principal] += ("fallback",)
         return all_perms
 
-    def get_acl(self, package):
-        """ Construct an ACL for a package """
+    def get_acl(self, package: str) -> List[Tuple[str, str, str]]:
+        """Construct an ACL for a package"""
         acl = []
         permissions = self.allowed_permissions(package)
-        for principal, perms in permissions.iteritems():
+        for principal, perms in permissions.items():
             for perm in perms:
                 acl.append((Allow, principal, perm))
         return acl
 
-    def has_permission(self, package, perm):
-        """ Check if this user has a permission for a package """
-        current_userid = self.request.userid
+    def has_permission(self, package: str, perm: str) -> bool:
+        """Check if this user has a permission for a package"""
+        current_userid = self.request.authenticated_userid
         if current_userid is not None and self.is_admin(current_userid):
             return True
 
         perms = self.allowed_permissions(package)
-        for principal in effective_principals(self.request):
+        for principal in self.user_principals(current_userid):
             if perm in perms.get(principal, []):
                 return True
         return False
 
-    def user_principals(self, username):
+    def user_principals(self, username: Optional[str]) -> List[str]:
         """
         Get a list of principals for a user
 
@@ -116,20 +206,22 @@ class IAccessBackend(object):
         principals : list
 
         """
-        principals = ['user:' + username, Everyone, Authenticated]
+        if username is None:
+            return [Everyone]
+        principals = ["user:" + username, Everyone, Authenticated]
         if self.is_admin(username):
-            principals.append('admin')
+            principals.append(Admin)
         for group in self.groups(username):
-            principals.append('group:' + group)
+            principals.append("group:" + group)
         return principals
 
-    def in_group(self, username, group):
+    def in_group(self, username: Optional[str], group: str) -> bool:
         """
         Find out if a user is in a group
 
         Parameters
         ----------
-        username : str
+        username : str, None
             Name of user. May be None for the anonymous user.
         group : str
             Name of the group. Supports 'everyone', 'authenticated', and
@@ -140,18 +232,18 @@ class IAccessBackend(object):
         member : bool
 
         """
-        if group in ('everyone', Everyone):
+        if group in ("everyone", Everyone):
             return True
         elif username is None:
             return False
-        elif group in ('authenticated', Authenticated):
+        elif group in ("authenticated", Authenticated):
             return True
-        elif group == 'admin' and self.is_admin(username):
+        elif group in ("admin", Admin) and self.is_admin(username):
             return True
         else:
             return group in self.groups(username)
 
-    def in_any_group(self, username, groups):
+    def in_any_group(self, username: str, groups: List[str]) -> bool:
         """
         Find out if a user is in any of a set of groups
 
@@ -170,13 +262,13 @@ class IAccessBackend(object):
         """
         return any((self.in_group(username, group) for group in groups))
 
-    def can_update_cache(self):
+    def can_update_cache(self) -> bool:
         """
         Return True if the user has permissions to update the pypi cache
         """
-        return self.in_any_group(self.request.userid, self.cache_update)
+        return self.in_any_group(self.request.authenticated_userid, self.cache_update)
 
-    def need_admin(self):
+    def need_admin(self) -> bool:
         """
         Find out if there are any admin users
 
@@ -191,7 +283,7 @@ class IAccessBackend(object):
         """
         return False
 
-    def allow_register(self):
+    def allow_register(self) -> bool:
         """
         Check if the backend allows registration
 
@@ -204,7 +296,20 @@ class IAccessBackend(object):
         """
         return False
 
-    def verify_user(self, username, password):
+    def allow_register_token(self) -> bool:
+        """
+        Check if the backend allows registration via tokens
+
+        This should only be overridden by mutable backends
+
+        Returns
+        -------
+        allow : bool
+
+        """
+        return False
+
+    def verify_user(self, username: str, password: str) -> bool:
         """
         Check the login credentials of a user
 
@@ -227,13 +332,13 @@ class IAccessBackend(object):
             user_data = self.user_data(username)
             if user_data is None:
                 return False
-        return bool(stored_pw and pwd_context.verify(password, stored_pw))
+        return bool(stored_pw and self.pwd_context.verify(password, stored_pw))
 
-    def _get_password_hash(self, username):
-        """ Get the stored password hash for a user """
+    def _get_password_hash(self, username: str) -> str:
+        """Get the stored password hash for a user"""
         raise NotImplementedError
 
-    def groups(self, username=None):
+    def groups(self, username: Optional[str] = None) -> List[str]:
         """
         Get a list of all groups
 
@@ -251,7 +356,7 @@ class IAccessBackend(object):
         """
         raise NotImplementedError
 
-    def group_members(self, group):
+    def group_members(self, group: str) -> List[str]:
         """
         Get a list of users that belong to a group
 
@@ -267,7 +372,7 @@ class IAccessBackend(object):
         """
         raise NotImplementedError
 
-    def is_admin(self, username):
+    def is_admin(self, username: str) -> bool:
         """
         Check if the user is an admin
 
@@ -282,63 +387,44 @@ class IAccessBackend(object):
         """
         raise NotImplementedError
 
-    def group_permissions(self, package, group=None):
+    def group_permissions(self, package: str) -> Dict[str, List[str]]:
         """
         Get a mapping of all groups to their permissions on a package
 
-        If a group is specified, just return the list of permissions for that
-        group
-
         Parameters
         ----------
         package : str
             The name of a python package
-        group : str, optional
-            The name of a single group the check
 
 
         Returns
         -------
         permissions : dict
-            If group is None, mapping of group name to a list of permissions
+            mapping of group name to a list of permissions
             (which can contain 'read' and/or 'write')
-        permissions : list
-            If group is not None, a list of permissions for that group
-
-        Notes
-        -----
-        You may specify special groups 'everyone' and/or 'authenticated', which
-        correspond to all users and all logged in users respectively.
 
         """
         raise NotImplementedError
 
-    def user_permissions(self, package, username=None):
+    def user_permissions(self, package: str) -> Dict[str, List[str]]:
         """
         Get a mapping of all users to their permissions for a package
 
-        If a username is specified, just return the list of permissions for
-        that user
-
         Parameters
         ----------
         package : str
             The name of a python package
-        username : str
-            The name of a single user the check
 
         Returns
         -------
         permissions : dict
             Mapping of username to a list of permissions (which can contain
             'read' and/or 'write')
-        permissions : list
-            If username is not None, a list of permissions for that user
 
         """
         raise NotImplementedError
 
-    def user_package_permissions(self, username):
+    def user_package_permissions(self, username: str) -> List[Dict[str, List[str]]]:
         """
         Get a list of all packages that a user has permissions on
 
@@ -355,7 +441,7 @@ class IAccessBackend(object):
         """
         raise NotImplementedError
 
-    def group_package_permissions(self, group):
+    def group_package_permissions(self, group: str) -> List[Dict[str, List[str]]]:
         """
         Get a list of all packages that a group has permissions on
 
@@ -389,7 +475,20 @@ class IAccessBackend(object):
         """
         raise NotImplementedError
 
-    def dump(self):
+    def check_health(self) -> Tuple[bool, str]:
+        """
+        Check the health of the access backend
+
+        Returns
+        -------
+        (healthy, status) : (bool, str)
+            Tuple that describes the health status and provides an optional
+            status message
+
+        """
+        return True, ""
+
+    def dump(self) -> Dict[str, Any]:
         """
         Dump all of the access control data to a universal format
 
@@ -399,39 +498,40 @@ class IAccessBackend(object):
 
         """
         from pypicloud import __version__
-        data = {}
-        data['allow_register'] = self.allow_register()
-        data['version'] = __version__
+
+        data = {}  # type: Dict[str, Any]
+        data["allow_register"] = self.allow_register()
+        data["version"] = __version__
 
         groups = self.groups()
         users = self.user_data()
         for user in users:
-            user['password'] = self._get_password_hash(user['username'])
+            user["password"] = self._get_password_hash(user["username"])
 
-        data['groups'] = {}
+        data["groups"] = {}
         packages = {
-            'users': defaultdict(dict),
-            'groups': defaultdict(dict),
-        }
+            "users": defaultdict(dict),
+            "groups": defaultdict(dict),
+        }  # type: Dict[str, Any]
         for group in groups:
-            data['groups'][group] = self.group_members(group)
+            data["groups"][group] = self.group_members(group)
             perms = self.group_package_permissions(group)
             for perm in perms:
-                package = perm['package']
-                packages['groups'][package][group] = perm['permissions']
+                package = perm["package"]
+                packages["groups"][package][group] = perm["permissions"]
 
         for user in users:
-            username = user['username']
+            username = user["username"]
             perms = self.user_package_permissions(username)
             for perm in perms:
-                package = perm['package']
-                packages['users'][package][username] = perm['permissions']
+                package = perm["package"]
+                packages["users"][package][username] = perm["permissions"]
 
         # Convert the defaultdict to a dict for easy serialization
-        packages['users'] = dict(packages['users'])
-        packages['groups'] = dict(packages['groups'])
-        data['users'] = users
-        data['packages'] = packages
+        packages["users"] = dict(packages["users"])
+        packages["groups"] = dict(packages["groups"])
+        data["users"] = users
+        data["packages"] = packages
         return data
 
     def load(self, data):
@@ -445,8 +545,10 @@ class IAccessBackend(object):
         necessary.
 
         """
-        raise TypeError("Access backend '%s' is not mutable and has no "
-                        "'load' implementation" % self.__class__.__name__)
+        raise TypeError(
+            "Access backend '%s' is not mutable and has no "
+            "'load' implementation" % self.__class__.__name__
+        )
 
 
 class IMutableAccessBackend(IAccessBackend):
@@ -455,18 +557,84 @@ class IMutableAccessBackend(IAccessBackend):
     Base class for access backends that can change user/group permissions
 
     """
+
     mutable = True
 
-    def need_admin(self):
+    def need_admin(self) -> bool:
         for user in self.user_data():
-            if user['admin']:
+            if user["admin"]:
                 return False
         return True
 
+    def get_signup_token(self, username: str) -> str:
+        """
+        Create a signup token
+
+        Parameters
+        ----------
+        username : str
+            The username to be created when this token is consumed
+
+        Returns
+        -------
+        token : str
+
+        """
+        msg, signature = self._hmac(username, time.time())
+        return msg + ":" + signature
+
+    def _hmac(self, username: str, timestamp: float) -> Tuple[str, str]:
+        """HMAC a username/expiration combo"""
+        if self.signing_key is None:
+            raise RuntimeError("auth.signing_key is not set!")
+        msg = "%s:%d" % (username, timestamp)
+        return (
+            msg,
+            hmac.new(
+                self.signing_key.encode("utf8"), msg.encode("utf8"), hashlib.sha256
+            ).hexdigest(),
+        )
+
+    def validate_signup_token(self, token: str) -> Optional[str]:
+        """
+        Validate a signup token
+
+        Parameters
+        ----------
+        token : str
+
+        Returns
+        -------
+        username : str or None
+            This will be None if the validation fails
+
+        """
+        if self.signing_key is None:
+            return None
+        pieces = token.split(":")
+        signature = pieces.pop()
+        username = pieces[0]
+        issued = int(pieces[1])
+        if issued + self.token_expiration < time.time():
+            return None
+        _, expected = self._hmac(username, issued)
+        if hasattr(hmac, "compare_digest"):
+            if not hmac.compare_digest(
+                signature.encode("utf-8"), expected.encode("utf-8")
+            ):
+                return None
+        else:
+            if signature != expected:
+                return None
+        return username
+
     def allow_register(self):
         raise NotImplementedError
 
-    def set_allow_register(self, allow):
+    def allow_register_token(self):
+        return self.signing_key is not None
+
+    def set_allow_register(self, allow: bool) -> None:
         """
         Allow or disallow user registration
 
@@ -477,7 +645,7 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def register(self, username, password):
+    def register(self, username: str, password: str) -> None:
         """
         Register a new user
 
@@ -490,10 +658,9 @@ class IMutableAccessBackend(IAccessBackend):
             This should be the plaintext password
 
         """
-        if self.allow_register():
-            self._register(username, pwd_context.encrypt(password))
+        self._register(username, self.pwd_context.hash(password))
 
-    def _register(self, username, password):
+    def _register(self, username: str, password: str) -> None:
         """
         Register a new user
 
@@ -508,7 +675,7 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def pending_users(self):
+    def pending_users(self) -> List[str]:
         """
         Retrieve a list of all users pending admin approval
 
@@ -520,7 +687,7 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def approve_user(self, username):
+    def approve_user(self, username: str) -> None:
         """
         Mark a user as approved by the admin
 
@@ -531,7 +698,7 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def edit_user_password(self, username, password):
+    def edit_user_password(self, username: str, password: str) -> None:
         """
         Change a user's password
 
@@ -541,9 +708,9 @@ class IMutableAccessBackend(IAccessBackend):
         password : str
 
         """
-        self._set_password_hash(username, pwd_context.encrypt(password))
+        self._set_password_hash(username, self.pwd_context.hash(password))
 
-    def _set_password_hash(self, username, password_hash):
+    def _set_password_hash(self, username: str, password_hash: str) -> None:
         """
         Change a user's password
 
@@ -556,7 +723,7 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def delete_user(self, username):
+    def delete_user(self, username: str) -> None:
         """
         Delete a user
 
@@ -567,7 +734,7 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def set_user_admin(self, username, admin):
+    def set_user_admin(self, username: str, admin: bool) -> None:
         """
         Grant or revoke admin permissions for a user
 
@@ -580,7 +747,7 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def edit_user_group(self, username, group, add):
+    def edit_user_group(self, username: str, group: str, add: bool) -> None:
         """
         Add or remove a user to/from a group
 
@@ -594,7 +761,7 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def create_group(self, group):
+    def create_group(self, group: str) -> None:
         """
         Create a new group
 
@@ -605,7 +772,7 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def delete_group(self, group):
+    def delete_group(self, group: str) -> None:
         """
         Delete a group
 
@@ -616,13 +783,15 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def edit_user_permission(self, package, username, perm, add):
+    def edit_user_permission(
+        self, package_name: str, username: str, perm: Set[str], add: bool
+    ) -> None:
         """
         Grant or revoke a permission for a user on a package
 
         Parameters
         ----------
-        package : str
+        package_name : str
         username : str
         perm : {'read', 'write'}
         add : bool
@@ -631,13 +800,15 @@ class IMutableAccessBackend(IAccessBackend):
         """
         raise NotImplementedError
 
-    def edit_group_permission(self, package, group, perm, add):
+    def edit_group_permission(
+        self, package_name: str, group: str, perm: Set[str], add: bool
+    ) -> None:
         """
         Grant or revoke a permission for a group on a package
 
         Parameters
         ----------
-        package : str
+        package_name : str
         group : str
         perm : {'read', 'write'}
         add : bool
@@ -651,49 +822,43 @@ class IMutableAccessBackend(IAccessBackend):
         pending_users = []
         for username in self.pending_users():  # pylint: disable=E1101
             password = self._get_password_hash(username)
-            pending_users.append({
-                'username': username,
-                'password': password,
-            })
-        data['pending_users'] = pending_users
+            pending_users.append({"username": username, "password": password})
+        data["pending_users"] = pending_users
         return data
 
     def load(self, data):
-        # Have to temporarily set this as True for the load operation
-        self.set_allow_register(True)
         pending_users = set(self.pending_users())
 
         def user_exists(username):
-            """ Helper function that checks if a user already exists """
-            return (username in pending_users or
-                    self.user_data(username) is not None)
+            """Helper function that checks if a user already exists"""
+            return username in pending_users or self.user_data(username) is not None
 
-        for user in data['users']:
-            if not user_exists(user['username']):
-                self._register(user['username'], user['password'])
-                self.approve_user(user['username'])
-            self.set_user_admin(user['username'], user.get('admin', False))
+        for user in data["users"]:
+            if not user_exists(user["username"]):
+                self._register(user["username"], user["password"])
+                self.approve_user(user["username"])
+            self.set_user_admin(user["username"], user.get("admin", False))
 
-        for group, members in data['groups'].iteritems():
-            if len(self.group_members(group)) == 0:
+        for group, members in data["groups"].items():
+            if not self.group_members(group):
                 self.create_group(group)
             current_members = self.group_members(group)
             add_members = set(members) - set(current_members)
             for member in add_members:
                 self.edit_user_group(member, group, True)
 
-        for user in data.get('pending_users', []):
-            if not user_exists(user['username']):
-                self._register(user['username'], user['password'])
+        for user in data.get("pending_users", []):
+            if not user_exists(user["username"]):
+                self._register(user["username"], user["password"])
 
-        for package, groups in data['packages']['groups'].iteritems():
-            for group, permissions in groups.iteritems():
+        for package, groups in data["packages"]["groups"].items():
+            for group, permissions in groups.items():
                 for perm in permissions:
                     self.edit_group_permission(package, group, perm, True)
 
-        for package, users in data['packages']['users'].iteritems():
-            for user, permissions in users.iteritems():
+        for package, users in data["packages"]["users"].items():
+            for user, permissions in users.items():
                 for perm in permissions:
                     self.edit_user_permission(package, user, perm, True)
 
-        self.set_allow_register(data['allow_register'])
+        self.set_allow_register(data["allow_register"])
