@@ -1,13 +1,15 @@
 """ Tools and resources for traversal routing """
 import functools
+from typing import Any, Callable, Dict
 
 
 class IStaticResource(object):
 
-    """ Simple resource base class for static-mapping of paths """
-    __name__ = ''
+    """Simple resource base class for static-mapping of paths"""
+
+    __name__ = ""
     __parent__ = None
-    subobjects = {}
+    subobjects = {}  # type: Dict[str, Callable[[Any], Any]]
 
     def __init__(self, request):
         self.request = request
@@ -21,16 +23,19 @@ class IStaticResource(object):
 
 class IResourceFactory(object):
 
-    """ Resource that generates child resources from a factory """
-    __name__ = ''
+    """Resource that generates child resources from a factory"""
+
+    __name__ = ""
     __parent__ = None
-    __factory__ = lambda x: None
+
+    def __factory__(self, name):
+        return None
 
     def __init__(self, request):
         self.request = request
 
     def __getitem__(self, name):
-        child = self.__factory__(name)
+        child = self.__factory__(name)  # pylint: disable=E1128
         child.__name__ = name
         child.__parent__ = self
         return child
@@ -38,7 +43,7 @@ class IResourceFactory(object):
 
 class SimpleResource(object):
 
-    """ Resource for simple pip calls """
+    """Resource for simple pip calls"""
 
     def __init__(self, request):
         self.request = request
@@ -52,7 +57,7 @@ class SimpleResource(object):
 
 class SimplePackageResource(object):
 
-    """ Resource for requesting simple endpoint package versions """
+    """Resource for requesting simple endpoint package versions"""
 
     __parent__ = None
     __name__ = None
@@ -65,7 +70,7 @@ class SimplePackageResource(object):
 
 class APIPackagingResource(IResourceFactory):
 
-    """ Resource for api package queries """
+    """Resource for api package queries"""
 
     def __init__(self, request):
         super(APIPackagingResource, self).__init__(request)
@@ -74,19 +79,20 @@ class APIPackagingResource(IResourceFactory):
 
 class APIPackageResource(IResourceFactory):
 
-    """ Resource for requesting package versions """
+    """Resource for requesting package versions"""
 
     def __init__(self, request, name):
         super(APIPackageResource, self).__init__(request)
         self.name = name
         self.__factory__ = functools.partial(
-            APIPackageFileResource, self.request, self.name)
+            APIPackageFileResource, self.request, self.name
+        )
         self.__acl__ = request.access.get_acl(self.name)
 
 
 class APIPackageFileResource(object):
 
-    """ Resource for api endpoints dealing with a single package version """
+    """Resource for api endpoints dealing with a single package version"""
 
     __parent__ = None
     __name__ = None
@@ -99,27 +105,53 @@ class APIPackageFileResource(object):
 
 class APIResource(IStaticResource):
 
-    """ Resource for api calls """
-    subobjects = {
-        'package': APIPackagingResource,
-    }
+    """Resource for api calls"""
+
+    subobjects = {"package": APIPackagingResource}
 
 
 class AdminResource(IStaticResource):
 
-    """ Resource for admin calls """
+    """Resource for admin calls"""
+
+
+class PackagesResource(IStaticResource):
+
+    """Resource for cleaner buildout config"""
+
+
+class AccountResource(object):
+
+    """Resource for login/logout endpoints"""
+
+    __name__ = ""
+    __parent__ = None
+
+    def __init__(self, request):
+        self.request = request
 
 
 class Root(IStaticResource):
 
-    """ Root context for PyPI Cloud """
+    """Root context for PyPI Cloud"""
+
     subobjects = {
-        'api': APIResource,
-        'admin': AdminResource,
-        'simple': SimpleResource,
-        'pypi': SimpleResource,
+        "api": APIResource,
+        "admin": AdminResource,
+        "simple": SimpleResource,
+        "pypi": SimpleResource,
+        "packages": PackagesResource,
+        "acct": AccountResource,
     }
 
     def __init__(self, request):
         super(Root, self).__init__(request)
         self.__acl__ = request.access.ROOT_ACL
+
+    def __getitem__(self, name):
+        try:
+            return super().__getitem__(name)
+        except KeyError:
+            # Allow fetching packages from the root (e.g. https://pypiserver.com/mypackage)
+            # This is for feature parity with pypiserver (see https://github.com/stevearc/pypicloud/issues/305)
+            return SimplePackageResource(self.request, name)
