@@ -1,24 +1,28 @@
 """ Base class for storage backends """
+from typing import BinaryIO, List, Tuple, Type
+
+from pyramid.request import Request
+
 from pypicloud.models import Package
 
 
 class IStorage(object):
 
-    """ Base class for a backend that stores package files """
+    """Base class for a backend that stores package files"""
 
-    def __init__(self, request):
+    def __init__(self, request: Request):
         self.request = request
 
     @classmethod
     def configure(cls, settings):
-        """ Configure the storage method with app settings """
+        """Configure the storage method with app settings"""
         return {}
 
-    def list(self, factory=Package):
-        """ Return a list or generator of all packages """
+    def list(self, factory: Type[Package] = Package) -> List[Package]:
+        """Return a list or generator of all packages"""
         raise NotImplementedError
 
-    def get_url(self, package):
+    def get_url(self, package: Package) -> str:
         """
         Create or return an HTTP url for a package file
 
@@ -30,15 +34,16 @@ class IStorage(object):
         -------
         link : str
             Link to the location of this package file
-        changed : bool
-            If True, the package needs to be saved to the cache
 
         """
-        url = self.request.app_url('api/package', package.name,
-                                   package.filename)
-        return (url, False)
+        fragment = ""
+        if package.data.get("hash_sha256"):
+            fragment = "sha256=" + package.data["hash_sha256"]
+        return self.request.app_url(
+            "api", "package", package.name, package.filename, fragment=fragment
+        )
 
-    def download_response(self, package):
+    def download_response(self, package: Package):
         """
         Return a HTTP Response that will download this package
 
@@ -47,7 +52,7 @@ class IStorage(object):
         """
         raise NotImplementedError
 
-    def upload(self, package, data):
+    def upload(self, package: Package, datastream: BinaryIO) -> None:
         """
         Upload a package file to the storage backend
 
@@ -55,13 +60,13 @@ class IStorage(object):
         ----------
         package : :class:`~pypicloud.models.Package`
             The package metadata
-        data : file
+        datastream : file
             A file-like object that contains the package data
 
         """
         raise NotImplementedError
 
-    def delete(self, package):
+    def delete(self, package: Package) -> None:
         """
         Delete a package file
 
@@ -73,7 +78,7 @@ class IStorage(object):
         """
         raise NotImplementedError
 
-    def open(self, package):
+    def open(self, package: Package):
         """
         Get a buffer object that can read the package data
 
@@ -94,3 +99,16 @@ class IStorage(object):
 
         """
         raise NotImplementedError
+
+    def check_health(self) -> Tuple[bool, str]:
+        """
+        Check the health of the storage backend
+
+        Returns
+        -------
+        (healthy, status) : (bool, str)
+            Tuple that describes the health status and provides an optional
+            status message
+
+        """
+        return True, ""
